default_query = {
    "query": {
        "filtered_query": {
            "query": {
                "match_all_query": {}
            },
            "filter": {
                "bool_filter": {
                    "operator": "or",
                    "filters": []
                }
            }
        }
    }
}



import inspect
from datetime import datetime
from apiclient import (
    APIClient,
    paginated,
    retry_request
)

from sfcc_connector import api
from sfcc_connector.api import *

from .queries import default_query


def get_next_page(response, previous_page_params):
    return {
        "count": response["count"],
        "start": response["start"] + response["count"]
    }


# Extend the client for your API integration.
class APIConnector(APIClient):
    """ Definition of client integrating with
        SFCC via OCAPI.
    """

    def __init__(self, **args):
        if "config" in args.keys():
            # Setting up configuration
            cfg = args.pop("config")
            Endpoint.config = cfg

        super().__init__(**args)

        # Configuring all available endpoints
        endpoints = dict([(name, cls) for name, cls in api.__dict__.items() if isinstance(cls, type)])
        for key, endpoint in endpoints.items():
            # print("Entering endpoint {}".format(key))
            if key.endswith("Endpoint") and key != "Endpoint":
                # e.g. objs are as such 'Endpoint': <class 'src.connector.api.api.Endpoint'>
                cfg = Endpoint.config
                ocapi_v = cfg.ocapi_version
                base_url = cfg.base_url
                domain = cfg.base_domain
                site_id = cfg.site_id

                for name, value in inspect.getmembers(endpoint):
                    if name.startswith("_") or inspect.ismethod(value) or inspect.isfunction(value):
                        # Ignore any private or class attributes.
                        continue
                    new_value = str(value).lstrip("/")
                    # TODO consider using Template
                    resource = new_value.replace("{url}", base_url) \
                                        .replace("{ocapi_version}", ocapi_v) \
                                        .replace("{domain}", domain)
                    # FIXME ?
                    if key.startswith("Data"):
                        resource = resource.replace(site_id, "-")
                    setattr(endpoint, name, resource)

    @retry_request
    def authenticate(self) -> dict:
        """ Perform authentication at SFCC. When successful, returns valid token.

            Retry policy:
            * Every request responding w/ 5xx status code
            * Max of 5 minutes with exponential backoff strategy with a maximum wait time of 30 seconds.
            * Retrying will stop after 5 minutes
        """

        grant_type = {
            "grant_type": "urn:demandware:params:oauth:grant-type:client-id:dwsid:dwsecuretoken"
        }

        return self.post(AuthEndpoint.access_token.format(client_id=Endpoint.config.client_id), grant_type)

    @paginated(by_query_params=get_next_page)
    @retry_request
    def get_orders(self, begin_date: datetime, end_date: datetime, custom_query=dict(), custom_select=str()) -> dict:
        dt_format = "%Y-%m-%dT%H:%M:%S.000Z"

        if custom_query:
            query = custom_query

        else:
            query = default_query
            range_filters = [{
                "range_filter": {
                    "field": "creation_date",
                    "from": "{}".format(begin_date.strftime(dt_format)),
                    "to": "{}".format(end_date.strftime(dt_format))
                }
            }, {
                "range_filter": {
                    "field": "last_modified",
                    "from": "{}".format(begin_date.strftime(dt_format)),
                    "to": "{}".format(end_date.strftime(dt_format))
                }
            }]

            query["sorts"] = [{
                "field": "order_no",
                "sort_order": "asc"
            }]

            query["query"]["filtered_query"]["filter"]["bool_filter"]["filters"] = range_filters

        # To reduce amount of info retrieved at SFCC, use select as shown below
        if custom_select:
            query["select"] = custom_select
            # query["select"] = "(count,total,start,hits.(data.(order_no,creation_date)))"

        return self.post(ShopEndpoint.order_search, query)

    @retry_request
    def get_order(self, order: int) -> dict:
        return self.get(ShopEndpoint.order.format(order_no=order))

    @paginated(by_query_params=get_next_page)
    @retry_request
    def get_jobs(self, begin_date: datetime, end_date: datetime, custom_query=dict(), custom_select=str()) -> dict:
        dt_format = "%Y-%m-%dT%H:%M:%S.000Z"

        if custom_query:
            query = custom_query

        else:
            query = default_query
            range_filters = [{
                "range_filter": {
                    "field": "start_time",
                    "from": "{}".format(begin_date.strftime(dt_format)),
                    "to": "{}".format(end_date.strftime(dt_format))
                }
            }, {
                "range_filter": {
                    "field": "end_time",
                    "from": "{}".format(begin_date.strftime(dt_format)),
                    "to": "{}".format(end_date.strftime(dt_format))
                }
            }]

            query["sorts"] = [{
                "field": "id",
                "sort_order": "asc"
            }]

            query["query"]["filtered_query"]["filter"]["bool_filter"]["filters"] = range_filters

        if custom_select:
            query["select"] = custom_select

        return self.post(DataEndpoint.job_execution_search, query)

    #
    # Add here other API calls
    #



import base64
import re
from datetime import datetime, timedelta
from typing import Optional
from pydantic import BaseModel, validator, ValidationError

from apiclient import (
    HeaderAuthentication,
    JsonResponseHandler,
    JsonRequestFormatter
)

from .extensions import (
    LoginRequestFormatter,
    SFCCErrorHandler
)

from .client import APIConnector


class AppConfig(BaseModel):
    # TODO extend validation / constraints on these values
    domain: str
    ocapi_version: Optional[str] = "v21_3"
    site_id: Optional[str] = "-"
    use_ssl: Optional[bool] = False
    is_production: Optional[bool] = False
    username: str
    password: str
    client_id: str
    client_password: str

    @property
    def base_url(self):
        protocol = "https://" if self.use_ssl else "http://"
        site_id_url = "/s/{}".format(self.site_id)
        prod_url = "{}".format(site_id_url if not self.is_production else "")
        return "{}{}{}/dw".format(protocol, self.domain, prod_url)
        # http(s)://public_domain[/s/site_id]/dw/api_type/ PROD
        # http(s)://sub_domain.demandware.net/s/site_id/dw/api_type/ NOT-prod

    @property
    def base_domain(self):
        protocol = "https://" if self.use_ssl else "http://"
        return "{}{}".format(protocol, self.domain)

    @property
    def auth_header(self):
        ret_val = "{0}:{1}:{2}".format(
            self.username,
            self.password,
            self.client_password).encode('utf-8')
        return base64.b64encode(ret_val)

    @validator("ocapi_version")
    def validate_ocapi_version(cls, v):
        pattern = re.compile("^v[0-9]{1,2}_[0-9]{1,2}$")
        if not pattern.match(v):
            raise ValueError("Unknown format of OCAPI Version. Supported format \
                is v<year>_<consecutive number> (e.g. 'v13_1')")
        return v


class SFCCClient(object):
    """This class will handle authentication and API calls towards SFCC.

       Initialization may raise APIRequestError and other exception
    """
    @property
    def configuration(self):
        return self._configuration

    @configuration.setter
    def configuration(self, new_configuration):
        self._configuration = new_configuration

    @property
    def expiration(self):
        return self._expiration

    @expiration.setter
    def expiration(self, new_expiration):
        self._expiration = datetime.now() + timedelta(seconds=new_expiration)

    @property
    def token(self):
        return self._token

    @token.setter
    def token(self, new_token):
        self._token = new_token

    def authenticate(self):
        """ 1. Perform authentication against SFCC
            2. Set token and its expiration
        """
        # Instantiate a client for authentication
        oauth_client = APIConnector(
            config=self.configuration,
            authentication_method=HeaderAuthentication(
                token=self.configuration.auth_header.decode(),
                parameter="Authorization",
                scheme="Basic",
            ),
            response_handler=JsonResponseHandler,
            error_handler=SFCCErrorHandler,
            request_formatter=LoginRequestFormatter
        )

        # This call may raise APIRequestError
        resp = oauth_client.authenticate()
        self.token = resp["access_token"]
        self.expiration = resp["expires_in"]

        return resp

    def __init__(self, cfg: dict):
        try:
            self.configuration = AppConfig(**cfg)
        except ValidationError as e:
            raise ValueError("Configuration initialization failed with error {}".format(e))

    def __getattr__(self, attr):
        """ Route calls to the client object
        """
        # Instantiate a client for all other calls
        if hasattr(APIConnector, attr):
            def wrapper(*args, **kw):
                try:
                    if datetime.now() >= self.expiration:
                        print("Token expired. Authenticating before executing '{}'".format(attr))
                        self.authenticate()
                except AttributeError as e:
                    print("Token not found. Authenticating before executing '{}' [Error: {}]".format(attr, e))
                    self.authenticate()

                return getattr(
                    APIConnector(
                        authentication_method=HeaderAuthentication(token=self.token),
                        response_handler=JsonResponseHandler,
                        error_handler=SFCCErrorHandler,
                        request_formatter=JsonRequestFormatter
                    ), attr)(*args, **kw)
            return wrapper
        raise AttributeError(attr)

    def __format__(self, format_str):
        """ Format message to be printed out
        """
        if self.token and datetime.now() < self.expiration:
            return "Authenticated. Token {} is still valid".format(self.token)
        else:
            return "Token expired or not found. Please authenticate again."




from apiclient.utils.typing import (
    OptionalJsonType,
    OptionalStr,
)

from apiclient.request_formatters import BaseRequestFormatter

from apiclient.error_handlers import BaseErrorHandler
from apiclient import exceptions
from apiclient.response import Response


class LoginRequestFormatter(BaseRequestFormatter):
    """Format the outgoing data as x-www-form-urlencoded
    """

    content_type = "application/x-www-form-urlencoded"

    @classmethod
    def format(cls, data: OptionalJsonType) -> OptionalStr:
        if data:
            return "{}={}".format(list(data.keys())[0], list(data.values())[0])


class SFCCErrorHandler(BaseErrorHandler):
    """Format the error messages according to SFCC definition
    """

    @staticmethod
    def get_exception(response: Response) -> exceptions.APIRequestError:
        """Parses client errors to extract bad request reasons."""
        json = response.get_json()
        err_args = json["fault"]["arguments"]
        err_msg = json["fault"]["message"]
        err_type = json["fault"]["type"]
        error = "{0}: {1} - {2}".format(err_type, err_msg, err_args)

        if 400 <= response.get_status_code() < 500:
            # json = response.get_json()
            return exceptions.ClientError(error, response.get_status_code())

        return exceptions.APIRequestError(error, response.get_status_code())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from setuptools import setup
from pathlib import Path

project_dir = Path(__file__).parent

about = {}
with open(project_dir.joinpath("sfcc_connector", "__main__.py"), "r") as f:
    exec(f.read(), about)

setup(
    name = about["__title__"],
    version = about["__version__"],
    description = about["__description__"],
    # Use UTF-8 encoding for README even on Windows by using the encoding argument.
    long_description = project_dir.joinpath("README.md").read_text(encoding="utf-8"),
    long_description_content_type = "text/markdown",
    author = about["__author__"],
    author_email=about["__author_email__"],
    maintainer=about["__maintainer__"],
    maintainer_email=about["__maintainer_email__"],
    license=about["__license__"],
    url = about["__url__"],
    keywords = about["__keywords__"],
    packages = ["sfcc_connector",],
    install_requires = project_dir.joinpath("requirements.in").read_text().split("\n"),
    python_requires = ">=3.6",
    include_package_data = True,
    zip_safe = False,
    classifiers = [
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



__title__ = "sfcc connector"
__description__ = "Python connector for Salesforce Commerce Cloud via Open Commerce API"
__url__ = "https://github.com/edro15/sfcc-connector"
__version__ = "1.0.2"
__author__ = "Erica Pescio"
__author_email__ = "edro15@users.noreply.github.com"
__license__ = "MIT"
__maintainer__ = "Erica Pescio"
__maintainer_email__ = "edro15@users.noreply.github.com"
__keywords__ = ["python", "salesforce commerce cloud", "ocapi", "open commerce api", "sfcc"]
